-- Puffin: A Tojaso FowlPlay Project
-- Model component for monitoring spell casting for player and target, plus mirror timers

local _, Puffin = ...
local EF = Puffin.Error
local DB = Puffin.Database

local Model_Casting = {}
local unit_changed = {}
local unit_tracking = {}
local mirrorIcons = {}
local mirrorRecord = {}
local sentTime = nil
local latency = 0
local callback = nil
local callbackActive = false

local function CastingEvent(e, unit) if unit == "vehicle" then unit = "player" end if unit_tracking[unit] then unit_changed[unit] = true; Model_Casting.OnCastEvent() end end
local function CastingUnitEvent(e, unit, track) unit_tracking[unit] = track; if unit ~= "mirror" then unit_changed[unit] = true end end -- set when castbar enabled for a unit
local function CastingSentEvent(e, unit) if unit == "player" or unit == "vehicle" then sentTime = GetTime() end end -- used to measure latency
local function CastingStartEvent(e, unit) if (unit == "player" or unit == "vehicle") and sentTime then latency = GetTime() - sentTime end CastingEvent(e, unit) end
local function CastingTargetEvent(e) CastingEvent(e, "target") end
local function CastingFocusEvent(e) CastingEvent(e, "focus") end
local function CastingPetEvent(e) CastingEvent(e, "pet") end

-- Initialize by registering for relevant events
function Model_Casting.OnInitialize()
	DB.AllocateDatabase("casting", false)
	
	mirrorIcons.BREATH = GetSpellTexture(5697)
	mirrorIcons.DEATH = GetSpellTexture(98391)
	mirrorIcons.EXHAUSTION = GetSpellTexture(57723)
	mirrorIcons.FEIGNDEATH = GetSpellTexture(5384)
	table.wipe(mirrorRecord)
	DB.SetRedirect("casting", "mirror", mirrorRecord) -- only one mirror entry so can optimize

	Puffin.RegisterCallback("UNIT_SPELLCAST_START", CastingStartEvent) -- register for events that affect casting
	Puffin.RegisterCallback("UNIT_SPELLCAST_STOP", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_SENT", CastingSentEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_DELAYED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_FAILED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_INTERRUPTED", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_INTERRUPTIBLE", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_NOT_INTERRUPTIBLE", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_START", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_STOP", CastingEvent)
	Puffin.RegisterCallback("UNIT_SPELLCAST_CHANNEL_UPDATE", CastingEvent)
	Puffin.RegisterCallback("PLAYER_TARGET_CHANGED", CastingTargetEvent)
	Puffin.RegisterCallback("PLAYER_FOCUS_CHANGED", CastingFocusEvent)
	Puffin.RegisterCallback("UNIT_PET", CastingPetEvent)
	Puffin.RegisterCallback("Puffin_Casting_Unit", CastingUnitEvent)
	
	-- hide blizzard casting and mirror bars
	CastingBarFrame:UnregisterAllEvents() -- player castbar
	
	local m = _G["MirrorTimer1"]; if m then m:Hide(); m.Show = function() end end -- mirror bar 1-3
	m = _G["MirrorTimer2"]; if m then m:Hide(); m.Show = function() end end
	m = _G["MirrorTimer3"]; if m then m:Hide(); m.Show = function() end end
end

-- On update, manage database of active cooldowns in response to cooldown-related events
function Model_Casting.OnCastEvent()
	local colors = DB.GetRecord("defaults", "castbar")
	for unit, changed in pairs(unit_changed) do
		if changed and unit_tracking[unit] then
			unit_changed[unit] = false
			local checkUnit = unit
			if UnitHasVehicleUI("player") then
				if unit == "player" then checkUnit = "pet" elseif unit == "pet" then checkUnit = "player" end
			end
			local channel = false
			local name, _, label, icon, startTime, endTime, tradeSkill, _, interrupt = UnitCastingInfo(checkUnit)
			if not name then
				name, _, label, icon, startTime, endTime, tradeSkill, interrupt = UnitChannelInfo(checkUnit)
				channel = true; label = name
			end
			if name then
				DB.SetAttribute("casting", unit, "label", label)
				DB.SetAttribute("casting", unit, "icon", icon)
				DB.SetAttribute("casting", unit, "expire", endTime / 1000)
				DB.SetAttribute("casting", unit, "duration", (endTime - startTime) / 1000)
				DB.SetAttribute("casting", unit, "tradeSkill", tradeSkill)
				DB.SetAttribute("casting", unit, "interrupt", interrupt)
				DB.SetAttribute("casting", unit, "channel", channel)
				DB.SetAttribute("casting", unit, "color", channel and colors.channel or colors.cast)
				DB.SetAttribute("casting", unit, "background", interrupt and colors.interrupt or colors.background)
				if not channel and unit == "player" then DB.SetAttribute("casting", unit, "latency", latency) end
				-- Puffin.Debug("casting", unit, name, label, icon, (endTime - startTime) / 1000, endTime / 1000, tradeSkill, interrupt, channel, latency)
			else
				DB.ReleaseRecord("casting", unit)
			end
			Puffin.Callback("Puffin_Casting" .. unit)
		end
	end
end

-- Mirror bar events are not reliabe so check during updates
function Model_Casting.OnUpdate()
	local found, same
	if unit_tracking.mirror then
		for i = 1, MIRRORTIMER_NUMTIMERS do
			local timer, timeLeft, duration, fillRate, paused, label = GetMirrorTimerInfo(i)
			if duration and duration > 0 then
				timeLeft = GetMirrorTimerProgress(timer)
				if timeLeft and (duration > timeLeft) then
					local expire = (timeLeft / 1000) + GetTime()
					if timer == mirrorRecord.timer and (math.floor(expire * 100) == math.floor(mirrorRecord.expire * 100)) then -- check if same
						same = true
					else
						mirrorRecord.label = label
						mirrorRecord.icon = mirrorIcons[timer]
						mirrorRecord.expire = expire
						mirrorRecord.duration = duration / 1000
						mirrorRecord.fillRate = fillRate
						if fillRate > 1 then mirrorRecord.timeLeft = timeLeft / 1000 else mirrorRecord.timeLeft = nil end -- for rapid refill
						mirrorRecord.timer = timer
						mirrorRecord.channel = true
						local colors = DB.GetRecord("defaults", "castbar")
						mirrorRecord.color = colors.mirror
						mirrorRecord.background = colors.background
						-- Puffin.Debug("mirror", i, timer, label, duration / 1000, timeLeft / 1000, fillRate)
					end
					found = true
					break -- only one active mirror timer at a time
				end
			end
		end
	end
	if found then
		if not callbackActive then Puffin.ResumeCallback(callback); callbackActive = true end
	else
		if mirrorRecord.timer then table.wipe(mirrorRecord) else same = true end
		if callbackActive then Puffin.SuspendCallback(callback); callbackActive = false end
	end
	if not same then Puffin.Callback("Puffin_Castingmirror") end
end

-- On refresh, update progress for any active mirror castbar
function Model_Casting.OnRefresh()
	local fillRate = mirrorRecord.fillRate
	if fillRate and fillRate > 1 then
		local timeLeft = GetMirrorTimerProgress(mirrorRecord.timer)
		if not timeLeft then timeLeft = 0 end
		mirrorRecord.timeLeft = timeLeft / 1000
		Puffin.Callback("Puffin_Castingmirror")
	end
end

Puffin.RegisterCallback("Puffin_Initialize_Priority", Model_Casting.OnInitialize)
Puffin.RegisterCallback("Puffin_Update_Priority", Model_Casting.OnUpdate)
callback = Puffin.RegisterCallback("Puffin_Refresh_Priority", Model_Casting.OnRefresh, 0, false)
